// #####################################
// ### MRF: Markov random field model
// #####################################

// this is C code for fitting a MRF model for binary spatial data
// see Paciorek C.J., in press in Computational Statistics and Data Analysis, 2007: Computational techniques for spatial logistic regression with large data sets for more details
// the code is only minimally commented
// note that this requires the use of the GMRFLib C library
// code is modified from template code of H. Rue

#include <assert.h>
#include <stddef.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#if !defined(__FreeBSD__)
#include <malloc.h>
#endif

// requires gsl library for Gaussian CDF and inverse CDF calculations
#include <gsl.h>

/* 
   include the required header-file
 */
#include "/opt/GMRFLib/GMRFLib.h"


/* 
   some useful macros
 */
#define Calloc(n,type) (type *)calloc((size_t)(n),sizeof(type))
#define FREE(ptr) {if(ptr)free((void *)(ptr));ptr=NULL;}
#define MAX(a,b)  ((a) > (b) ? (a) : (b))
#define MIN(a,b)  ((a) < (b) ? (a) : (b))
#define SQR(x)    ((x)*(x))

#define STORE_INTERVAL 1
#define NITS 21000
#define NBURN 1000

/* 
   parameters in the gamma-prior for kappa
 */
#define PRIOR_A  1.0                            
#define PRIOR_B 0.01

/* 
   global variable G containing the data and the graph. gridObs contains the association between observations and grid cells, namely the number of observations in each cell and, if any, the indices indicating which observations are in the cell
 */
typedef struct
{
    GMRFLib_graph_tp *graph;
}
Global;
static Global G, gridObs;


double eu_gamma(double x, double a, double b)
{
    /* 
       return the unnormalized log-density of a gamma-variable with mean a/b.
    */
    return (a-1.0)*log(x)-x*b;
}


double truncN(double* w,double* x,int* y,int* mapping,double mu,int N){
     // samples a vector of truncated normals 

    double p,tmpx,dev;
    int mult,i;
    for(i=0;i<N;i++){
        mult=(y[i] == 1 ? -1 : 1);
        tmpx=mult*(mu+x[mapping[i]-1]);
        p=gsl_cdf_ugaussian_P(-tmpx);
        dev=GMRFLib_uniform()*p;
        w[i]= mult*(tmpx+gsl_cdf_ugaussian_Pinv(dev));
    }
    return 0;
}

double Qfunc(int node, int nnode, char *arg)
{
    /* 
       return Q_ij, i=node, j=nnode, for the prior

       \pi(x) \sim \exp(-\frac{1}{2} \sum_{i\sim j} (x_i-x_j)^2)
    */
    
    GMRFLib_graph_tp *g;
    char **args;
    double kappa;
    
    args  = (char **)arg;                         /* the required variables come */
    g     = (GMRFLib_graph_tp *)args[0];          /* through the arg-pointer */
    kappa = *((double *)args[1]);

    return kappa*(node == nnode ? g->nnbs[node] : -1.0);
}

int main(int argc, char **argv)
{

    int seed, i, j,k, counter,num;
    int N, buf;
    int Nf, K, tmp;
    double fac,*bvec, *cvec, *x,*xmean, *xnew, *d, kappa, kappa_new, lacc, eprob=0.0, timeref;
    double *w;
    double mu, mn, LP;
    int *y, *filled, *mapping;
    GMRFLib_problem_tp *Gproblem;
    char *arguments_new[2], *arguments_old[2], *arguments_loglik[2];
    GMRFLib_constr_tp *constr = NULL;
    GMRFLib_blockupdate_param_tp *blockupdate_param;
    GMRFLib_optimize_param_tp *optimize_param;
        
    printf("Usage: %s SEED\n", *argv);
    assert(argc > 1);
    
    /* 
       get the seed and init the random generator in GMRFLib
    */
    seed = atoi(argv[1]);
    printf("seed %d\n", seed);
 
    GMRFLib_uniform_init(seed);

    char* itsfile="name of output file for iteration values"
    char* pmfile="name of output file for posterior mean estimates"

// the following input files overlap and some could be eliminated if more of the manipulation were done here - essentially all that are needed is the observations and the vector grid cells to which the observations belong
    char* yfile="name of input file of y observations"  // 
    char* mappingfile="name of input file of which grid cell each observation is in" // cells should be referred to as 1,...,K*K; file should start with N, the number of observations
    char* filledfile="name of input file of a sorted vector of which grid cells have observations" // again, use 1,...,K*K ; file should start with Nf, the number of filled cells
    char* gridobsfile="name of input file associating each grid cell with its observations"  // file should start with the number of observations in the first cell, followed by the indices (numbered 1, ... ,n) of the observations in that cell, then the number of observations in the second cell, followed by the indices, etc.; cells that do not have any observations should be indicated with a zero followed by no indices
    
    K=32;

   FILE *fp;
    //   get graph and data
   
    GMRFLib_make_lattice_graph(&G.graph,K,K,1,1,0);
   
    fp = fopen(mappingfile, "r");
    fscanf(fp,"%ld\n",&buf);
    N=buf;
    mapping=Calloc(N,int);
    for(i=0;i<N;i++){
       if (fscanf(fp, "%ld \n", &buf ) != 1) abort();
       mapping[i]=buf;
    }
    fclose(fp);

    fp = fopen(yfile, "r");
    y=Calloc(N,int);
     for(i=0;i<N;i++){
       if (fscanf(fp, "%ld \n", &buf ) != 1) abort();
       y[i]=buf;
     }
    fclose(fp);


    cvec=Calloc(G.graph->n, double);
    for(i=0;i<G.graph->n;i++){
        cvec[i]=0.0;
       }
    for(i=0;i<N;i++){
        cvec[mapping[i]-1]+=1.0;
    }

    fp = fopen(filledfile, "r");
    fscanf(fp,"%ld\n",&buf);
    Nf=buf;
    filled=Calloc(Nf,int);
    for(i=0;i<Nf;i++){
       if (fscanf(fp, "%ld \n", &buf ) != 1) abort();
       filled[i]=buf;
    }
    fclose(fp);

    GMRFLib_make_empty_graph(&gridObs.graph);
    fp = fopen(gridObsfile, "r");
    gridObs.graph->nobs=Calloc(G.graph->n,int);
    gridObs.graph->obs=Calloc(G.graph->n,int*);
    for(i=0;i<G.graph->n;i++){
    fscanf(fp,"%ld\n",&buf);
    gridObs.graph->nobs[i]=buf;
    if(gridObs.graph->nobs[i]>0){
      gridObs.graph->obs[i]=Calloc(gridObs.graph->nobs[i],int);
       for(k=0;k<gridObs.graph->nobs[i];k++){
         fscanf(fp,"%ld\n",&buf);
         gridObs.graph->obs[i][k]=buf;
       }
    }
    }

    fp=fopen(itsfile,"w");
    fprintf(fp," ");
    // setup space

  
    
    w    = Calloc(N, double);
    x    = Calloc(G.graph->n, double);
    xmean    = Calloc(G.graph->n, double);
    xnew = Calloc(G.graph->n, double);
    bvec = Calloc(G.graph->n, double);


//       inital values
  
    kappa =  10.0;  // 10                     


    for(i=0;i<G.graph->n;i++)
    {
         x[i]=GMRFLib_stdnormal();
         bvec[i]=0.0;
         xmean[i]=0.0;
    }

    mn=0.0;
    for(i=0;i<N;i++){
         mn=mn+y[i];
    }
    mn=mn/(double)N;
    mu=exp(mn)/(1+exp(mn));

    /* 
       the scale-factor for proposing a new value for kappa
    */
    fac = 3.5;

 
    /* 
       setup arguments to the log-likelihood function, which needs
       data (y and E), the graph, and kappa.
    */
    arguments_new[0]    = (char *)G.graph;
    arguments_new[1]    = (char *)&kappa_new;
    arguments_old[0]    = (char *)G.graph;
    arguments_old[1]    = (char *)&kappa;

    /* 
       choose whether or not the constraint \sum x_i = 0  is required
    */

    if (1)
    {
        printf("constraining sum to zero.\n");
        GMRFLib_make_empty_constr(&constr);
        constr->nc = 1;
        constr->a_matrix = Calloc(G.graph->n, double);
        for(i=0;i<G.graph->n;i++) constr->a_matrix[i] = 1.0;
        constr->e_vector = Calloc(1, double);
        constr->e_vector[0] = 0.0;
        GMRFLib_prepare_constr(constr, G.graph,  0);
    }
    else
        constr = NULL;

    /* 
       get default options for the blockupdate-routine and the
       optimizing-routine which is (optionally) used in the
       blockupdate-routine. now it is easy to change options if required.
    */
    GMRFLib_default_blockupdate_param(&blockupdate_param);
    GMRFLib_default_optimize_param(&optimize_param);
    
    timeref= GMRFLib_ctime();
    for(counter=1;counter<=NITS;counter++)
    {

        truncN(w,x,y,mapping,mu,N);  // sample w from truncated Normal
  
        mn=0.0;
        for(i=0;i<N;i++){
           mn+=w[i]-x[mapping[i]-1];
        }
        mu=(mn/(double)N)+sqrt(1.0/(double)N)*GMRFLib_stdnormal();
        // sample mu from conditional under flat prior

        // bvec is vector with one value for each grid point, containing the sum of the w values (offset by mu) for the observations falling into the grid cell (0 for grid points with no observations) - this is required for the GMRF process sampling below
        for(i=0;i<Nf;i++){
           bvec[filled[i]-1]=0.0;
           for(j=0;j<gridObs.graph->nobs[filled[i]-1];j++){   
              bvec[filled[i]-1]+=w[(gridObs.graph->obs[filled[i]-1][j])-1]-mu;   
           }
        } 

        /* 
           propose a new value for kappa
	*/
        kappa_new = kappa * GMRFLib_scale_proposal(fac);
  
        /* 
           conditional on kappa_new, propose a new value for the GMRF x
	*/

        GMRFLib_blockupdate(
	    &lacc,
	    xnew, x,
	    bvec,bvec,  /* NULL, NULL,  */         /* zero vectors */
	    cvec,cvec,  /* NULL, NULL*/           /* zero vectors */
	    NULL, NULL,           /* zero vectors */
	    NULL, NULL,  /* d, d, */
	    NULL, NULL, /*loglik, (char *)arguments_loglik, */
            NULL,NULL, /*	    loglik, (char *)arguments_loglik, */
	    NULL,                 /* none fixed values, update all */
	    G.graph,
	    Qfunc, (char *)arguments_new,  
	    Qfunc, (char *)arguments_old,
	    NULL, NULL,           /* default values used here */
	    NULL, NULL,           /* default values used here */
	    constr, constr, 
	    optimize_param, blockupdate_param);
        
       /* 
           the blockupdate-routine does not compute the normalization
	   constant terms for the GMRF, nor the prior for kappa
	*/

        lacc += (G.graph->n-1.0)/2.0*log(kappa_new/kappa) 
            + eu_gamma(kappa_new, PRIOR_A, PRIOR_B)
            - eu_gamma(kappa, PRIOR_A, PRIOR_B);

        /* 
           standard accept/reject
	*/
        if (GMRFLib_uniform()<exp(MIN(0.0, lacc)))
        {
            kappa = kappa_new;
            memcpy(x, xnew, G.graph->n*sizeof(double));
        }


        /* 
           that is it !
	*/
        if(counter>NBURN){
            for(i=0;i<G.graph->n;i++)
        {       
         xmean[i]=(xmean[i]*(double)(counter-NBURN-1)+mu+x[i])/(double)(counter-NBURN);
         }
         if((counter-NBURN)%STORE_INTERVAL==0){

         fp=fopen(itsfile,"a");
         fprintf(fp,"%.4f ",kappa);

             for(i=0;i<G.graph->n;i++){
             fprintf(fp,"%.4f ",x[i]+mu);
             }     
             fprintf(fp,"\n");             
             fclose(fp) ;
         }
         }        


        eprob += exp(MIN(0.0, lacc));       
        printf("lacc= %.8f  E(accept_prob)= %.4f iter/src= %.4f  kappa= %.4f x1= %.4f\n",
	       lacc, eprob/counter, 1./((GMRFLib_ctime()-timeref)/counter), kappa, x[965]);
    }
    fp = fopen(pmfile, "w");

    for(i=0;i<G.graph->n;i++){
       fprintf(fp,"%.4f ",xmean[i]);
       }  
    fclose(fp) ;



   return 0;

}


